# -*- coding: utf-8 -*-
from pathlib import Path
from typing import Any, Dict, List, NoReturn, TypeVar, Union, cast

import structlog
from jinja2 import Template as JinjaTemplate
from jinja2 import TemplateNotFound as JinjaTemplateNotFound
from orjson import (
    OPT_INDENT_2,
    OPT_NON_STR_KEYS,
    OPT_OMIT_MICROSECONDS,
    OPT_SERIALIZE_NUMPY,
    dumps,
)
from pydantic import DirectoryPath
from pydantic_openapi_schema.v3_1_0.open_api import OpenAPI
from ruamel.yaml import YAML
from starlite import (
    CORSConfig,
    Provide,
    Request,
    Response,
    Starlite,
    State,
    StaticFilesConfig,
    TemplateConfig,
)
from starlite.enums import MediaType, OpenAPIMediaType
from starlite.exceptions import ImproperlyConfiguredException, TemplateNotFoundException
from starlite.status_codes import HTTP_204_NO_CONTENT, HTTP_304_NOT_MODIFIED
from starlite.template import TemplateEngineProtocol
from starlite.types import (
    ControllerRouterHandler,
    ExceptionHandlersMap,
)
from starlite.utils import create_exception_response

from kiara.context import Kiara
from kiara.interfaces.python_api import KiaraAPI
from kiara.models import KiaraModel
from kiara.registries.templates import TemplateRegistry
from kiara.utils import is_debug, is_develop
from kiara_plugin.service.defaults import KIARA_SERVICE_RESOURCES_FOLDER
from kiara_plugin.service.openapi.controllers.jobs import JobControllerJson
from kiara_plugin.service.openapi.controllers.modules import ModuleControllerJson
from kiara_plugin.service.openapi.controllers.operations import (
    OperationControllerJson,
)
from kiara_plugin.service.openapi.controllers.pipeline import PipelineControllerJson
from kiara_plugin.service.openapi.controllers.render import RenderControllerJson
from kiara_plugin.service.openapi.controllers.values import (
    ValueControllerJson,
)
from kiara_plugin.service.openapi.controllers.workflows import WorkflowControllerJson

T = TypeVar("T")


logger = structlog.getLogger()
yaml = YAML(typ="safe")


class KiaraModelResponse(Response):
    @classmethod
    def serializer(cls, value: Any) -> Dict[str, Any]:
        if isinstance(value, KiaraModel):
            return value.dict()
        return super().serializer(value)

    def render(self, content: Any) -> bytes:
        """
        Handles the rendering of content T into a bytes string.
        Args:
            content: An arbitrary value of type T

        Returns:
            An encoded bytes string
        """
        try:
            if (
                content is None
                or content is NoReturn
                and (
                    self.status_code < 100
                    or self.status_code in {HTTP_204_NO_CONTENT, HTTP_304_NOT_MODIFIED}
                )
            ):
                return b""
            if self.media_type == MediaType.JSON:
                return dumps(
                    content,
                    default=self.serializer,
                    option=OPT_SERIALIZE_NUMPY
                    | OPT_OMIT_MICROSECONDS
                    | OPT_NON_STR_KEYS,
                )
            if isinstance(content, OpenAPI):
                content_dict = content.dict(by_alias=True, exclude_none=True)
                if self.media_type == OpenAPIMediaType.OPENAPI_YAML:
                    encoded = yaml.dump(content_dict).encode("utf-8")
                    return cast("bytes", encoded)
                return dumps(
                    content_dict,
                    option=OPT_INDENT_2 | OPT_OMIT_MICROSECONDS | OPT_NON_STR_KEYS,
                )
            return super().render(content)
        except (AttributeError, ValueError, TypeError) as e:
            raise ImproperlyConfiguredException(
                "Unable to serialize response content"
            ) from e


def logging_exception_handler(request: Request, exc: Exception) -> Response:
    """
    Logs exception and returns appropriate response.

    Parameters
    ----------
    request : Request
        The request that caused the exception.
    exc :
        The exception caught by the Starlite exception handling middleware and passed to the
        callback.

    Returns
    -------
    Response
    """
    logger.error("Application Exception")
    return create_exception_response(exc)


# def http_exception_handler(_: Request, exc: Exception) -> Response:
#     """Default handler for exceptions subclassed from HTTPException"""
#     status_code = HTTP_500_INTERNAL_SERVER_ERROR
#     detail = ""
#     if hasattr(exc, "detail"):
#         detail = exc.detail
#     if hasattr(exc, "status_code"):
#         status_code = exc.status_code
#     return Response(
#         media_type=MediaType.TEXT,
#         content=detail,
#         status_code=status_code,
#     )
# def http_exception_handler(request: Request, exc: HTTPException):
#     return JSONResponse(
#         {"detail": exc.detail}, status_code=exc.status_code, headers=exc.headers
#     )


# def custom_exception_handler(request: Request, exc: Exception):
#     model = InternalErrorModel.from_exception(exc)
#     return JSONResponse({"detail": model.dict()}, status_code=model.status)


class KiaraOpenAPIService:
    def __init__(self, kiara_api: KiaraAPI):

        self._kiara_api: KiaraAPI = kiara_api
        self._app: Union[Starlite, None] = None
        self._resources_base: Path = Path(KIARA_SERVICE_RESOURCES_FOLDER)

    def app(self) -> Starlite:
        if self._app is not None:
            return self._app

        from starlite import Router

        value_router = Router(path="/data", route_handlers=[ValueControllerJson])
        module_router = Router(path="/modules", route_handlers=[ModuleControllerJson])
        operation_router = Router(
            path="/operations", route_handlers=[OperationControllerJson]
        )
        job_router = Router(path="/jobs", route_handlers=[JobControllerJson])
        render_router = Router(path="/render", route_handlers=[RenderControllerJson])
        workflow_router = Router(
            path="/workflows", route_handlers=[WorkflowControllerJson]
        )
        pipeline_router = Router(
            path="/pipelines", route_handlers=[PipelineControllerJson]
        )

        # info_router_html = Router(
        #     path="/html/info", route_handlers=[OperationControllerHtml]
        # )
        # Router(path="/html/values", route_handlers=[ValueControllerHtmx])
        # Router(path="/html/operations", route_handlers=[OperationControllerHtmx])

        route_handlers: List[ControllerRouterHandler] = []
        route_handlers.append(value_router)
        route_handlers.append(module_router)
        route_handlers.append(operation_router)
        route_handlers.append(job_router)
        route_handlers.append(render_router)
        route_handlers.append(workflow_router)
        route_handlers.append(pipeline_router)

        # route_handlers.append(value_router_htmx)
        # route_handlers.append(operation_router_htmx)

        static_dir = self._resources_base / "static"

        static_file_config = [
            StaticFilesConfig(directories=[static_dir], path="/static")
        ]
        self._template_registry: TemplateRegistry = TemplateRegistry()

        environment = self._template_registry.environment

        class KiaraTemplateEngine(TemplateEngineProtocol[JinjaTemplate]):
            """Template engine using the default kiara template registry."""

            def __init__(
                self, directory: Union[DirectoryPath, List[DirectoryPath]]
            ) -> None:
                super().__init__(directory=directory)
                self.engine = environment

            def get_template(self, name: str) -> JinjaTemplate:
                """Loads the template with the name and returns it."""
                try:
                    return self.engine.get_template(name=name)
                except JinjaTemplateNotFound as exc:
                    raise TemplateNotFoundException(template_name=name) from exc

        def engine_callback(jinja_engine: KiaraTemplateEngine) -> KiaraTemplateEngine:
            jinja_engine.engine.globals["kiara_api"] = self._kiara_api
            return jinja_engine

        template_config: TemplateConfig = TemplateConfig(
            directory=[], engine=KiaraTemplateEngine, engine_callback=engine_callback
        )

        debug = is_debug() or is_develop()

        cors_config = CORSConfig()
        exception_handlers: ExceptionHandlersMap = {}
        # exception_handlers[HTTPException] = http_exception_handler
        # exception_handlers[Exception] = custom_exception_handler
        # if is_debug() or is_develop():
        #     exception_handlers[HTTPException] = logging_exception_handler
        # exception_handlers[Exception] = http_exception_handler

        async def get_kiara_context(state: State) -> Kiara:
            if not hasattr(state, "kiara"):
                state.kiara = self._kiara_api.context
            return cast(Kiara, state.kiara)

        async def get_kiara_api(state: State) -> KiaraAPI:
            if not hasattr(state, "kiara_api"):
                state.kiara_api = self._kiara_api
            return cast(KiaraAPI, state.kiara_api)

        async def get_template_registry(state: State) -> TemplateRegistry:
            if not hasattr(state, "template_registry"):
                state.template_registry = self._template_registry
            return cast(TemplateRegistry, self._template_registry)

        dependencies = {
            "kiara": Provide(get_kiara_context),
            "kiara_api": Provide(get_kiara_api),
            "template_registry": Provide(get_template_registry),
        }

        self._app = Starlite(
            route_handlers=route_handlers,
            dependencies=dependencies,
            static_files_config=static_file_config,
            template_config=template_config,
            debug=debug,
            cors_config=cors_config,
            exception_handlers=exception_handlers,
            response_class=KiaraModelResponse,
        )
        return self._app  # type: ignore



# -*- coding: utf-8 -*-
from kiara.interfaces.python_api.models.job import JobTest

"""Auto-generated tests that use job descriptions in the 'examples/jobs' folder and run them.

To test against the outputs of those jobs, add files into subfolders that are called the same as the job (minus the file extension), under the `tests/job_tests` folder.

To test values directly, add a file called `outputs.json` or `outputs.yaml` into that folder, containing a 'dict'
data structure with the value attribute to test as key and the expected value as value.

Most likely you will want to test against a value property, which would be done like so (in yaml):

```yaml
network_data::properties::metadata.graph_properties::number_of_self_loops: 1
```

The format is:
```yaml
<output field name>::properties::<property name>::[<property attribute>]::[<more_sub_keys>]: <expected value>
```

In case of scalars, you can also test against the value directly:
```
output_string::data: "some string"
```

If the results are too complex to test against in this way, you can also write Python code. Add a file with a random
name and a `.py` extension (output.py is a good choice if you only have one). In that file, each function will be run
against the job results. You can control which arguments will be passed to the function by naming the arguments:

- `kiara_api`: a kiara api instance will be passed in
- 'outputs`: the whole result of the job will be passed in (of type `ValueMap`)
- the field name of the value you are interested in (e.g. `table`, `network_data`, depends on the job)

Specifying any other argument name will throw an error.
"""


def test_job_desc(example_job_test: JobTest):

    example_job_test.run_tests()



# -*- coding: utf-8 -*-
from typing import Any, Dict, List, Mapping, Union

from pydantic import BaseModel, Field
from starlite import Controller

from kiara.api import KiaraAPI, ValueSchema
from kiara.models.module.operation import Operation
from kiara.models.rendering import RenderValueResult
from kiara_plugin.service.openapi.controllers import get, post


class InputsValidationData(BaseModel):

    inputs: Mapping[str, Any] = Field(description="The provided inputs.")
    inputs_schema: Mapping[str, ValueSchema] = Field(description="The inputs schemas.")


class RenderControllerJson(Controller):
    path = "/"

    @get(
        path="/create_render_manifest/{data_type:str}",
        api_func=KiaraAPI.assemble_render_pipeline,
    )
    async def create_render_manifest(
        self, kiara_api: KiaraAPI, data_type: str
    ) -> Operation:
        """Create a render manifest for the specified data type."""

        filters = ["select_columns"]
        operation = kiara_api.assemble_render_pipeline(
            data_type=data_type, target_format="html", filters=filters
        )
        return operation

    @post(path="/value/{value:str}/{target_format:str}", api_func=KiaraAPI.render_value)
    async def render_data(
        self,
        kiara_api: KiaraAPI,
        value: str,
        target_format: str = "html",
        data: Union[None, Dict[str, Any]] = None,
    ) -> RenderValueResult:
        """Queue a render job for the specified value id or alias.

        Arguments:
            value: the value id or alias
            target_format: the render format
            data: (optional) target & data type specific render configuration

        Returns:
            the render result
        """

        # filters = ["select_columns", "drop_columns"]
        filters: List[str] = []
        v = kiara_api.get_value(value)
        result = kiara_api.render_value(
            value=v, target_format=target_format, filters=filters, render_config=data
        )
        return result

    @post(
        path="/value_info/{value:str}/{target_format:str}",
        summary="Render value info as HTML",
    )
    async def render_operation_info(
        self,
        kiara_api: KiaraAPI,
        value: str,
        target_format: str = "html",
        data: Union[Dict[str, Any], None] = None,
    ) -> str:

        print(f"RENDER VALUE INFO REQUEST: {value}")

        value_info = kiara_api.retrieve_value_info(value)
        html = value_info.create_html()
        return html



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for kiara_plugin.language_processing.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""
# import pytest


import os
import tempfile
import uuid
from pathlib import Path

import pytest

from kiara.context import KiaraConfig
from kiara.interfaces.python_api import KiaraAPI
from kiara.interfaces.python_api.models.job import JobTest
from kiara.utils.testing import get_tests_for_job, list_job_descs

ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
JOBS_FOLDER = Path(os.path.join(ROOT_DIR, "examples", "jobs"))


def create_temp_dir():
    session_id = str(uuid.uuid4())
    TEMP_DIR = Path(os.path.join(tempfile.gettempdir(), "kiara_tests"))

    instance_path = os.path.join(
        TEMP_DIR.resolve().absolute(), f"instance_{session_id}"
    )
    return instance_path


@pytest.fixture
def kiara_api() -> KiaraAPI:

    instance_path = create_temp_dir()
    kc = KiaraConfig.create_in_folder(instance_path)
    api = KiaraAPI(kc)
    return api


@pytest.fixture(params=list_job_descs(JOBS_FOLDER))
def example_job_test(request, kiara_api) -> JobTest:

    job_tests_folder = Path(os.path.join(ROOT_DIR, "tests", "job_tests"))

    job_desc = request.param
    tests = get_tests_for_job(
        job_alias=job_desc.job_alias, job_tests_folder=job_tests_folder
    )

    job_test = JobTest(kiara_api=kiara_api, job_desc=job_desc, tests=tests)
    return job_test


@pytest.fixture
def example_data_folder() -> Path:
    return Path(os.path.join(ROOT_DIR, "examples", "data"))


@pytest.fixture
def example_pipelines_folder() -> Path:
    return Path(os.path.join(ROOT_DIR, "examples", "pipelines"))


@pytest.fixture()
def tests_resources_folder() -> Path:
    return Path(os.path.join(ROOT_DIR, "tests"))



# -*- coding: utf-8 -*-
import os

import mkdocs_gen_files
from kiara.context import Kiara

kiara = Kiara.instance()

modules_file_path = os.path.join("modules_list.md")
modules_page_content = """# Available module types

This page contains a list of all available *Kiara* module types, and their details.

!!! note
The formatting here will be improved later on, for now this should be enough to get the important details of each module type.

"""

BASE_PACKAGE = "kiara_plugin.service"


for module_type in kiara.module_mgmt.find_modules_for_package(
    BASE_PACKAGE, include_pipelines=False
).keys():

    if module_type == "pipeline":
        continue

    modules_page_content = modules_page_content + f"## ``{module_type}``\n\n"
    modules_page_content = (
        modules_page_content
        + "```\n{{ get_module_info('"
        + module_type
        + "') }}\n```\n\n"
    )

with mkdocs_gen_files.open(modules_file_path, "w") as f:
    f.write(modules_page_content)

pipelines_file_path = os.path.join("pipelines_list.md")
pipelines_page_content = """# Available pipeline module types

This page contains a list of all available *Kiara* pipeline module types, and their details.

!!! note
The formatting here will be improved later on, for now this should be enough to get the important details of each module type.

"""

for module_type in kiara.module_mgmt.find_modules_for_package(
    BASE_PACKAGE, include_core_modules=False
):

    if module_type == "pipeline":
        continue

    pipelines_page_content = pipelines_page_content + f"## ``{module_type}``\n\n"
    pipelines_page_content = (
        pipelines_page_content
        + "```\n{{ get_module_info('"
        + module_type
        + "') }}\n```\n\n"
    )

with mkdocs_gen_files.open(pipelines_file_path, "w") as f:
    f.write(pipelines_page_content)



# -*- coding: utf-8 -*-
from typing import Dict, List, Union

from pydantic import BaseModel, Extra, Field
from starlite import (
    Controller,
)

from kiara.api import KiaraAPI
from kiara.interfaces.python_api import OperationInfo
from kiara_plugin.service.openapi.controllers import get, post


class OperationRequest(BaseModel):
    element_id: str = Field(description="The id of the element to be created.")
    operation_id: str = Field(description="The id of the operation.")


class OperationRunRequest(BaseModel):
    class Config:
        extra = Extra.allow

    element_id: str = Field(description="The id of the element to be created.")
    operation_id: str = Field(description="The id of the operation.")


class MonitorJobRequest(BaseModel):

    element_id: str = Field(description="The id of the result element.")
    job_id: str = Field(description="The id of the job to monitor.")


class OperationMatcher(BaseModel):

    filters: List[str] = Field(
        description="The (optional) filter strings, an operation must match all of them to be included in the result.",
        default_factory=list,
    )
    include_internal: bool = Field(
        description="Whether to include internal operations in the result.",
        default=False,
    )
    python_package: Union[str, None] = Field(
        description="If specified, only operations that are contained in this Python package are returned.",
        default=None,
    )


class OperationControllerJson(Controller):
    path = "/"

    @post(path="/", api_func=KiaraAPI.retrieve_operations_info)
    async def list_operations(
        self, kiara_api: KiaraAPI, data: OperationMatcher
    ) -> Dict[str, OperationInfo]:

        filters = data.filters
        include_internal = data.include_internal

        if data.python_package is not None:
            python_packages = [data.python_package]
        else:
            python_packages = None

        operations = kiara_api.retrieve_operations_info(
            *filters, include_internal=include_internal, python_packages=python_packages
        )
        return operations.item_infos  # type: ignore

    @post(path="/ids", api_func=KiaraAPI.list_operation_ids)
    async def list_operation_ids(
        self, kiara_api: KiaraAPI, data: OperationMatcher
    ) -> List[str]:
        """List the ids of all available operations."""

        filters = data.filters
        include_internal = data.include_internal

        if data.python_package is not None:
            python_packages = [data.python_package]
        else:
            python_packages = None

        operation_ids = kiara_api.list_operation_ids(
            filter=filters,
            include_internal=include_internal,
            python_packages=python_packages,
        )
        return operation_ids

    @get(path="/{operation_id:str}", api_func=KiaraAPI.retrieve_operation_info)
    async def get_operation_info(
        self, kiara_api: KiaraAPI, operation_id: str
    ) -> OperationInfo:

        op = kiara_api.retrieve_operation_info(operation=operation_id)
        return op


# class OperationControllerHtmx(Controller):
#     path = "/"
#
#     @get(path="/", media_type=MediaType.HTML)
#     async def get_root_page(self, kiara: Kiara) -> Template:
#
#         print("OPERATION ROOT REQUEST")
#         return Template(
#             name="kiara_plugin.service/operations/index.html", context={"kiara": kiara}
#         )
#
#     @post(path="/operation_info", media_type=MediaType.HTML)
#     async def render_operation_info(
#         self,
#         kiara_api: KiaraAPI,
#         data: OperationRequest = Body(media_type=RequestEncodingType.URL_ENCODED),
#     ) -> Template:
#
#         print(f"Operation info request: {data}")
#
#         op_info = kiara_api.retrieve_operation_info(operation=data.operation_id)
#
#         return Template(
#             name="kiara_plugin.service/operations/operation_view.html",
#             context={"element_id": data.element_id, "operation_info": op_info},
#         )
#
#     @post(path="/inputs_form", media_type=MediaType.HTML)
#     async def get_input_form(
#         self,
#         kiara_api: KiaraAPI,
#         template_registry: TemplateRegistry,
#         data: OperationRequest = Body(media_type=RequestEncodingType.URL_ENCODED),
#     ) -> Template:
#
#         print(f"INPUTS FORM REQUEST: {data}")
#         op = kiara_api.get_operation(data.operation_id)
#
#         fields = {}
#         for field_name, schema in op.inputs_schema.items():
#
#             data_type_cls = kiara_api.context.type_registry.get_data_type_cls(
#                 type_name=schema.type
#             )
#
#             template_name = f"kiara_plugin.service/values/inputs/{schema.type}.html"
#             if template_name not in template_registry.template_names:
#                 data_type_instance = data_type_cls(**schema.type_config)
#                 if data_type_instance.characteristics.is_scalar:
#                     template_name = (
#                         "kiara_plugin.service/values/inputs/generic-scalar.html"
#                     )
#                 else:
#                     template_name = "kiara_plugin.service/values/inputs/generic.html"
#             try:
#                 template = template_registry.get_template(template_name)
#                 rendered = template.render(
#                     field_name=field_name,
#                     data_type=schema.type,
#                     desc=schema.doc.description,
#                     doc=schema.doc.doc,
#                 )
#             except Exception as e:
#                 import traceback
#
#                 traceback.print_exc()
#                 rendered = str(e)
#
#             fields[field_name] = rendered
#
#         return Template(
#             name="kiara_plugin.service/values/value_inputs_form.html",
#             context={"fields": fields},
#         )
#
#     @post(path="/queue_job", media_type=MediaType.HTML)
#     async def queue_job(
#         self,
#         kiara_api: KiaraAPI,
#         data: OperationRunRequest = Body(media_type=RequestEncodingType.URL_ENCODED),
#     ) -> Union[Template, str]:
#
#         print(f"RUN REQUEST: {data.dict()}")
#
#         try:
#             inputs = data.dict()
#             operation_id = inputs.pop("operation_id")
#             element_id = inputs.pop("element_id")
#
#             # because html forms don't send values for unchecked boxes
#             op = kiara_api.get_operation(operation_id)
#             for field_name, schema in op.inputs_schema.items():
#                 if field_name not in inputs.keys():
#                     if schema.type == "boolean":
#                         inputs[field_name] = False
#
#             job_id = kiara_api.queue_job(operation=operation_id, inputs=inputs)
#
#             job = kiara_api.get_job(job_id=job_id)
#             return Template(
#                 name="kiara_plugin.service/jobs/job_monitor.html",
#                 context={"job": job, "element_id": element_id},
#             )
#
#         except Exception as e:
#             import traceback
#
#             traceback.print_exc()
#             return f"<div>Can't submit job: {e}"
#
#     @post(path="/monitor_job", media_type=MediaType.HTML)
#     async def monitor_job(
#         self,
#         kiara_api: KiaraAPI,
#         data: MonitorJobRequest = Body(media_type=RequestEncodingType.URL_ENCODED),
#     ) -> Template:
#
#         print(f"MONITOR REQUEST: {data.dict()}")
#
#         job_id = uuid.UUID(data.job_id)
#         job = kiara_api.get_job(job_id=job_id)
#
#         if job.finished is None:
#             return Template(
#                 name="kiara_plugin.service/jobs/job_monitor.html",
#                 context={"job": job, "element_id": data.element_id},
#             )
#         else:
#             if job.status == JobStatus.SUCCESS:
#                 results: Union[Dict, ValueMap] = kiara_api.get_job_result(job_id)
#                 error = None
#             else:
#                 results = {}
#                 error = job.error
#             return Template(
#                 name="kiara_plugin.service/jobs/job_finished.html",
#                 context={
#                     "job": job,
#                     "element_id": data.element_id,
#                     "results": results,
#                     "error": error,
#                 },
#             )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `kiara_plugin.service` package."""

import pytest  # noqa

import kiara_plugin.service


def test_assert():

    assert kiara_plugin.service.get_version() is not None



# -*- coding: utf-8 -*-
from typing import Any, Mapping

from pydantic import BaseModel, Extra, Field
from starlite import Controller

from kiara.api import KiaraAPI
from kiara.models.module.jobs import ActiveJob
from kiara_plugin.service.openapi.controllers import get, post


class RunJobRequest(BaseModel):
    class Config:
        extra = Extra.allow

    operation_id: str = Field(description="The id of the operation or module.")
    operation_config: Mapping[str, Any] = Field(
        description="The configuration of the operation.", default_factory=dict
    )
    inputs: Mapping[str, Any] = Field(description="The job input values.")


class JobInfoRequest(BaseModel):

    job_id: str = Field(description="The id of the job to monitor.")


class JobControllerJson(Controller):
    path = "/"

    @post(path="/queue_job", api_func=KiaraAPI.queue_job)
    async def queue_job(self, kiara_api: KiaraAPI, data: RunJobRequest) -> ActiveJob:

        print(f"JOB RUN REQUEST: {data.dict()}")

        try:
            operation_id = data.operation_id
            if not data.operation_config:
                job_id = kiara_api.queue_job(operation=operation_id, inputs=data.inputs)
            else:
                manifest = kiara_api.context.create_manifest(
                    module_or_operation=operation_id, config=data.operation_config
                )
                job_id = kiara_api.queue_job(operation=manifest, inputs=data.inputs)

            job = kiara_api.get_job(job_id=job_id)
            return job

        except Exception as e:
            import traceback

            traceback.print_exc()
            raise e

    @get(path="/monitor_job/{job_id:str}", api_func=KiaraAPI.get_job)
    async def monitor_job(self, kiara_api: KiaraAPI, job_id: str) -> ActiveJob:

        print(f"MONITOR REQUEST: {job_id}")

        job = kiara_api.get_job(job_id=job_id)

        return job



# -*- coding: utf-8 -*-
from typing import List

from pydantic import BaseModel, Field
from starlite import Controller

from kiara.api import KiaraAPI
from kiara.interfaces.python_api.models.info import PipelineStructureInfo
from kiara.utils.pipelines import get_pipeline_config
from kiara_plugin.service.openapi.controllers import get


class PipelineMatcher(BaseModel):

    filters: List[str] = Field(
        description="The (optional) filter strings, a pipeline must match all of them to be included in the result.",
        default_factory=list,
    )


class PipelineControllerJson(Controller):
    path = "/"

    @get(path="/structure/{pipeline:str}", api_func=get_pipeline_config)
    async def get_pipeline_structure(
        self, kiara_api: KiaraAPI, pipeline: str
    ) -> PipelineStructureInfo:

        print(f"PIPELINE: {pipeline}")
        pipeline_config = get_pipeline_config(pipeline=pipeline)
        info = PipelineStructureInfo.create_from_instance(
            kiara=kiara_api.context, instance=pipeline_config.structure
        )
        return info

    @get(path="/list", api_func=get_pipeline_config)
    async def list_pipelines(self, kiara_api: KiaraAPI) -> List[str]:

        pipelines = kiara_api.list_operations(operation_types="pipeline")
        print(pipelines)
        return list(pipelines.keys())



# -*- coding: utf-8 -*-
from typing import Any, Dict, Union

from pydantic import BaseModel, Extra, Field
from starlite import Controller, MediaType, get, post

from kiara.context import Kiara


class OperationControllerHtml(Controller):
    path = "/operation"

    @get(path="/{operation_name:str}", media_type=MediaType.HTML)
    def get_operation(self, operation_name: str, kiara: Kiara) -> str:

        op_info = kiara.context_info.operations.type_infos.get(operation_name)  # type: ignore
        return op_info.create_html()

    @post("/search_operation", media_type=MediaType.HTML)
    def search_operation(self, search_term: str, kiara: Kiara) -> str:

        for op in kiara.context_info.operations:
            pass

        return "xxx"


class DataTypeRequest(BaseModel):

    data_type: str = Field(description="The data type.", default="any")


class RenderRequest(BaseModel):
    class Config:
        extra = Extra.allow

    target_id: str = Field(description="The id of the target element.")
    field_name: str = Field(description="The field name.")
    render_conf: Union[Dict[str, Any], None] = Field(
        description="The scene render config.", default=None
    )


class DataTypeModel(BaseModel):

    field_name: str = Field(description="The field name.")
    data_type: str = Field(description="The data type.")
    type_config: Dict[str, str] = Field(
        description="The (optional) data type configuration.", default_factory=dict
    )



# -*- coding: utf-8 -*-
from typing import Dict, List, Union

from pydantic import BaseModel, Field
from starlite import Controller

from kiara.api import KiaraAPI
from kiara.models.workflow import WorkflowInfo
from kiara_plugin.service.openapi.controllers import get, post


class WorkflowMatcher(BaseModel):

    filters: List[str] = Field(
        description="The (optional) filter strings, an operation must match all of them to be included in the result.",
        default_factory=list,
    )


class WorkflowControllerJson(Controller):
    path = "/"

    @post(path="/ids", api_func=KiaraAPI.retrieve_workflows_info)
    async def list_workflows(
        self, kiara_api: KiaraAPI, data: Union[WorkflowMatcher, None] = None
    ) -> Dict[str, WorkflowInfo]:

        # if data is None:
        #     filters: List[str] = []
        # else:
        #     filters = data.filters

        result = kiara_api.retrieve_workflows_info().item_infos
        return result  # type: ignore

    @post(path="/aliases", api_func=KiaraAPI.list_workflow_alias_names)
    async def list_workflow_aliases(
        self, kiara_api: KiaraAPI, data: Union[WorkflowMatcher, None] = None
    ) -> List[str]:

        # if data is None:
        #     filters: List[str] = []
        # else:
        #     filters = data.filters

        result = kiara_api.list_workflow_alias_names()
        return result

    @get(
        path="/workflow_info/{workflow: str}", api_func=KiaraAPI.retrieve_workflow_info
    )
    async def get_workflow_info(
        self, kiara_api: KiaraAPI, workflow: str
    ) -> WorkflowInfo:

        print(f"INFO: {workflow}")
        workflow_info = kiara_api.retrieve_workflow_info(workflow=workflow)
        return workflow_info



# -*- coding: utf-8 -*-
from typing import Callable, Tuple, Union

import docstring_parser
from docstring_parser import DocstringStyle
from starlite import get as starlite_get
from starlite import post as starlite_post


def extract_doc(func: Callable) -> Tuple[Union[str, None], Union[str, None]]:

    doc = func.__doc__
    if not doc:
        return None, None

    docstring = docstring_parser.parse(doc, DocstringStyle.GOOGLE)
    summary = docstring.short_description
    description = docstring.long_description
    return summary, description


def get(*args, **kwargs) -> Callable:
    api_func = kwargs.pop("api_func", None)
    if api_func:
        summary, description = extract_doc(api_func)
        if summary:
            kwargs["summary"] = summary
        if description:
            kwargs["description"] = description
    return starlite_get(*args, **kwargs)


def post(*args, **kwargs) -> Callable:
    api_func = kwargs.pop("api_func", None)
    if api_func:
        kwargs["summary"], kwargs["description"] = extract_doc(api_func)
    return starlite_post(*args, **kwargs)



# -*- coding: utf-8 -*-

from pydantic import Field

from kiara.api import KiaraModule, KiaraModuleConfig, ValueMap, ValueMapSchema


class ExampleModuleConfig(KiaraModuleConfig):

    separator: str = Field(
        description="The seperator between the two strings.", default=" - "
    )


class ExampleModule(KiaraModule):
    """A very simple example module; concatenate two strings.

    The purpose of this modules is to show the main elements of a [`KiaraModule`][kiara.modules.KiaraModule]:

    - ***the (optional) configuration class***: must inherit from [`KiaraModuleConfig`][kiara.modules.KiaraModuleConfig], and the config class must be set as the `_config_cls` attribute
         on the `KiaraModule` class. Configuration values can be retrieved via the [`self.get_config_value(key)`][kiara.modules.KiaraModule.get_config_value] method
    - ***the inputs description***: must return a dictionary, containing the input name(s) as keys, and another dictionary containing type_name information
         and documentation about the input data as value
    - ***the outputs description***: must return a dictionary, containing the output name(s) as keys, and another dictionary containing type_name information
         and documentation about the output data as value
    - ***the ``process`` method***: this is where the actual work gets done. Input data can be accessed via ``inputs.get_value_data(key)``, results
         can be set with the ``outputs.set_value(key, value)`` method

    Example:

        This example module can be tested on the commandline with the ``kiara run`` command:

        ```
        kiara run core_types.example text_1="xxx" text_2="yyy"
        ```
    """

    _config_cls = ExampleModuleConfig
    _module_type_name = "service.example"

    def create_inputs_schema(
        self,
    ) -> ValueMapSchema:

        inputs = {
            "text_1": {"type": "string", "doc": "The first text."},
            "text_2": {"type": "string", "doc": "The second text."},
        }

        return inputs

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        outputs = {
            "text": {
                "type": "string",
                "doc": "The concatenated text.",
            }
        }
        return outputs

    def process(self, inputs: ValueMap, outputs: ValueMap) -> None:

        separator = self.get_config_value("separator")

        text_1 = inputs.get_value_data("text_1")
        text_2 = inputs.get_value_data("text_2")

        result = text_1 + separator + text_2
        outputs.set_value("text", result)



# -*- coding: utf-8 -*-
from typing import Dict, List, Union

from pydantic import BaseModel, Field
from starlite import (
    Controller,
)

from kiara.api import KiaraAPI
from kiara.interfaces.python_api import ModuleTypeInfo
from kiara_plugin.service.openapi.controllers import get, post

# class OperationRequest(BaseModel):
#     element_id: str = Field(description="The id of the element to be created.")
#     operation_id: str = Field(description="The id of the operation.")


class ModuleMatcher(BaseModel):

    filters: List[str] = Field(
        description="The (optional) filter strings, a module must match all of them to be included in the result.",
        default_factory=list,
    )
    python_package: Union[str, None] = Field(
        description="If specified, only modules that are contained in this Python package are returned.",
        default=None,
    )


class ModuleControllerJson(Controller):
    path = "/"

    @post(path="/", api_func=KiaraAPI.retrieve_module_types_info)
    async def list_module_types(
        self, kiara_api: KiaraAPI, data: ModuleMatcher
    ) -> Dict[str, ModuleTypeInfo]:

        filters = data.filters
        python_package = data.python_package

        module_types = kiara_api.retrieve_module_types_info(
            *filters, python_package=python_package
        )
        return module_types.item_infos  # type: ignore

    @get(path="/type_names", api_func=KiaraAPI.list_module_type_names)
    async def list_module_type_names(self, kiara_api: KiaraAPI) -> List[str]:
        """List the ids of all available operations."""

        module_names = kiara_api.list_module_type_names()
        return module_names

    @get(path="/{module_type_name:str}", api_func=KiaraAPI.retrieve_module_type_info)
    async def get_module_type_info(
        self, kiara_api: KiaraAPI, module_type_name: str
    ) -> ModuleTypeInfo:

        module = kiara_api.retrieve_module_type_info(module_type=module_type_name)
        return module



# -*- coding: utf-8 -*-
import os
import sys

from appdirs import AppDirs

kiara_html_app_dirs = AppDirs("kiara-service", "frkl")

if not hasattr(sys, "_MEIPASS"):
    KIARA_MODULE_BASE_FOLDER = os.path.dirname(__file__)
    """Marker to indicate the base folder for the `kiara_streamlit` module."""
else:
    KIARA_MODULE_BASE_FOLDER = os.path.join(sys._MEIPASS, "kiara")  # type: ignore
    """Marker to indicate the base folder for the `kiara_streamlit` module."""

KIARA_SERVICE_RESOURCES_FOLDER = os.path.join(KIARA_MODULE_BASE_FOLDER, "resources")
"""Default resources folder for this package."""



# -*- coding: utf-8 -*-
import uuid
from typing import Any, Dict, List, Mapping, Union

from networkx import DiGraph
from networkx.readwrite import json_graph
from pydantic import BaseModel, Field
from starlite import (
    Controller,
)

from kiara.api import Kiara, KiaraAPI, Value, ValueSchema
from kiara.exceptions import InvalidValuesException
from kiara.interfaces.python_api import ValueInfo, ValuesInfo
from kiara.models.values.matchers import ValueMatcher
from kiara.models.values.value import SerializedData
from kiara_plugin.service.openapi.controllers import get, post


class InputsValidationData(BaseModel):

    inputs: Mapping[str, Any] = Field(description="The provided inputs.")
    inputs_schema: Mapping[str, ValueSchema] = Field(description="The inputs schemas.")


class ValueControllerJson(Controller):
    path = "/"

    @get(path="/ids", api_func=KiaraAPI.list_value_ids)
    async def list_value_ids(self, kiara_api: KiaraAPI) -> List[uuid.UUID]:

        result = kiara_api.list_value_ids()
        return result

    @get(path="/value_info/{value: str}", api_func=KiaraAPI.retrieve_value_info)
    async def get_value_info(self, kiara_api: KiaraAPI, value: str) -> ValueInfo:

        value_info = kiara_api.retrieve_value_info(value=value)
        return value_info

    # @post(path="/values", api_func=KiaraAPI.retrieve_values_info)
    # async def find_values(
    #     self, kiara_api: KiaraAPI, data: ValueMatcher
    # ) -> Dict[str, ValueInfo]:
    #
    #     matcher_data = data.dict()
    #
    #     result = kiara_api.retrieve_values_info(**matcher_data).item_infos
    #     return result  # type: ignore

    @post(path="/values_info", api_func=KiaraAPI.retrieve_values_info)
    async def get_values_info(
        self, kiara_api: KiaraAPI, data: ValueMatcher
    ) -> Dict[str, ValueInfo]:

        matcher_data = data.dict()

        result = kiara_api.retrieve_values_info(**matcher_data)
        return result.item_infos  # type: ignore

    @get(path="/type/{data_type:str}/values", api_func=KiaraAPI.list_values)
    async def find_values_of_type(
        self, kiara_api: KiaraAPI, data_type: str
    ) -> Dict[str, Value]:

        matcher = ValueMatcher(data_types=[data_type])

        result = kiara_api.list_values(**matcher.dict())
        return result  # type: ignore
        # return {str(k): v for k, v in result.items()}

    @get(
        path="/type/{data_type:str}/values_info",
        summary="List values info of specific data type.",
    )
    async def find_values_info_of_type(
        self, kiara_api: KiaraAPI, data_type: str
    ) -> ValuesInfo:

        matcher = ValueMatcher(data_types=[data_type])

        result = kiara_api.retrieve_values_info(**matcher.dict())
        return result

    @post(path="/alias_names", api_func=KiaraAPI.list_alias_names)
    async def list_alias_names(
        self, kiara_api: KiaraAPI, data: ValueMatcher
    ) -> List[str]:

        matcher_data = data.dict()
        result = kiara_api.list_alias_names(**matcher_data)
        return result

    @post(path="/aliases", api_func=KiaraAPI.list_aliases)
    async def list_aliases(
        self, kiara_api: KiaraAPI, data: ValueMatcher
    ) -> Dict[str, Value]:

        matcher_data = data.dict()

        result = kiara_api.list_aliases(**matcher_data)
        return result  # type: ignore

    @post(path="/aliases_info", api_func=KiaraAPI.retrieve_aliases_info)
    async def list_aliases_info(
        self, kiara_api: KiaraAPI, data: ValueMatcher
    ) -> Dict[str, ValueInfo]:

        if data is None:
            matcher_data = {}
        else:
            matcher_data = data.dict()

        result = kiara_api.retrieve_aliases_info(**matcher_data)
        return result.item_infos  # type: ignore

    @get(path="/type/{data_type:str}/aliases", api_func=KiaraAPI.list_aliases)
    async def find_value_aliases_of_type(
        self, kiara_api: KiaraAPI, data_type: str
    ) -> Dict[str, Value]:

        matcher = ValueMatcher(data_types=[data_type], has_alias=True)

        result = kiara_api.list_aliases(**matcher.dict())
        return result  # type: ignore

    @get(path="/type/{data_type:str}/alias_names", api_func=KiaraAPI.list_alias_names)
    async def find_value_aliase_names_of_type(
        self, kiara_api: KiaraAPI, data_type: str
    ) -> List[str]:
        matcher = ValueMatcher(data_types=[data_type], has_alias=True)

        result = kiara_api.list_alias_names(**matcher.dict())
        return result

    @get(
        path="/type/{data_type:str}/aliases_info",
        api_func=KiaraAPI.retrieve_aliases_info,
    )
    async def find_value_aliases_info_of_type(
        self, kiara_api: KiaraAPI, data_type: str
    ) -> ValuesInfo:

        matcher = ValueMatcher(data_types=[data_type], has_alias=True)

        result = kiara_api.retrieve_aliases_info(**matcher.dict())
        return result

    @get(
        path="/serialized/{value:uuid}",
        summary="Retrieve the serialized form of the values data.",
    )
    async def retrieve_data(
        self, kiara_api: KiaraAPI, value: Union[str, uuid.UUID]
    ) -> SerializedData:

        _value = kiara_api.get_value(value)
        return _value.serialized_data

    async def filter_data(self, kiara: Kiara, value):
        raise NotImplementedError()

    @post(path="/validate/inputs", summary="Validate inputs against a schema.")
    async def validate_inputs(
        self, kiara_api: KiaraAPI, data: InputsValidationData
    ) -> Dict[str, str]:

        print("VALIDATE REQUEST")
        try:
            value_map = kiara_api.context.data_registry.create_valuemap(
                data=data.inputs, schema=data.inputs_schema
            )
            return value_map.check_invalid()
        except InvalidValuesException as ive:
            return dict(ive.invalid_inputs)

    @get(path="/lineage/{value:str}", summary="Retrieve the lineage data for a value.")
    async def get_value_lineage(
        self, kiara_api: KiaraAPI, value: str
    ) -> Dict[str, Any]:

        print(f"LINEAGE REQUEST: {value}")
        _value = kiara_api.get_value(value=value)
        try:
            graph: DiGraph = _value.lineage.module_graph
            result = json_graph.node_link_data(graph)
            return result
        except Exception as e:
            import traceback

            traceback.print_exc()
            raise e


# class ValueControllerHtmx(Controller):
#     path = "/"
#
#     @get(path="/", media_type=MediaType.HTML)
#     async def get_root_page(self, kiara: Kiara) -> Template:
#
#         return Template(
#             name="kiara_plugin.service/values/index.html", context={"kiara": kiara}
#         )
#
#     # @get(path="/values/aliases", media_type=MediaType.HTML)
#     # def get_alias_select_box(self, kiara: Kiara) -> Template:
#     #     return Template(name="kiara_plugin.service/values/alias_select.html", context={"kiara": kiara})
#
#     @post(path="/select", media_type=MediaType.HTML)
#     async def get_value_select(
#         self,
#         kiara_api: KiaraAPI,
#         data: DataTypeRequest = Body(media_type=RequestEncodingType.URL_ENCODED),
#     ) -> Template:
#
#         if data and data.data_type:
#             data_types = [data.data_type]
#         else:
#             data_types = []
#
#         print(f"DATA_SELECT: {data_types}")
#
#         if data and data.data_type:
#             data_types = [data.data_type]
#         else:
#             data_types = []
#
#         return Template(
#             name="kiara_plugin.service/values/value_select.html",
#             context={"data_types": data_types, "field_name": "__no_field_name__"},
#         )
#
#     @post(path="/render", media_type=MediaType.HTML)
#     async def render_value(
#         self,
#         kiara_api: KiaraAPI,
#         data: RenderRequest = Body(media_type=RequestEncodingType.URL_ENCODED),
#     ) -> Template:
#
#         print(f"RENDER REQUEST: {data}")
#         try:
#
#             if not hasattr(data, data.field_name):
#                 raise Exception(
#                     f"Request is missing the value attribute '{data.field_name}'."
#                 )
#
#             value_id = getattr(data, data.field_name)
#
#             value = kiara_api.get_value(value=value_id)
#
#             print("-------")
#             print(value)
#             render_conf = data.render_conf
#             if render_conf is None:
#                 render_conf = {}
#
#             print(render_conf)
#
#             render_result = kiara_api.render_value(
#                 value=value,
#                 target_format=["html", "string"],
#                 render_config=render_conf,
#             )
#         except Exception as e:
#             import traceback
#
#             traceback.print_exc()
#             raise e
#
#         return Template(
#             name="kiara_plugin.service/values/value_view.html",
#             context={
#                 "element_id": data.target_id,
#                 "render_value_result": render_result,
#                 "value_id": str(value.value_id),
#                 "field_name": data.field_name,
#             },
#         )
#
#     @post(path="/input_widget", media_type=MediaType.HTML)
#     async def get_input_element_for_type(
#         self,
#         kiara_api: KiaraAPI,
#         template_registry: TemplateRegistry,
#         data: DataTypeModel = Body(media_type=RequestEncodingType.URL_ENCODED),
#     ) -> str:
#
#         print(f"INPUT FIELD REQUEST: {data.data_type}")
#
#         data_type_cls = kiara_api.context.type_registry.get_data_type_cls(
#             type_name=data.data_type
#         )
#         data_type_instance = data_type_cls(**data.type_config)
#
#         alias_map = kiara_api.list_aliases(data_types=[data.data_type])
#
#         try:
#             template = template_registry.get_template(
#                 f"kiara_plugin.service/values/inputs/{data.data_type}.html"
#             )
#             rendered = template.render(
#                 data_type_instance=data_type_instance,
#                 alias_map=alias_map,
#                 data_type_name=data.data_type,
#                 field_name=data.field_name,
#             )
#         except Exception as e:
#             rendered = str(e)
#
#         return rendered



# -*- coding: utf-8 -*-

#  Copyright (c) 2021, University of Luxembourg / DHARPA project
#  Copyright (c) 2021, Markus Binsteiner
#
#  Mozilla Public License, version 2.0 (see LICENSE or https://www.mozilla.org/en-US/MPL/2.0/)

"""Web-service related subcommands for the cli."""
import typing

import rich_click as click

if typing.TYPE_CHECKING:
    from kiara.api import KiaraAPI


@click.group()
@click.pass_context
def service(ctx):
    """(Web-)service-related sub-commands."""


@service.command()
@click.option(
    "--host", help="The host to bind to.", required=False, default="localhost"
)
@click.option("--port", "-p", help="The port to bind to.", required=False, default=8080)
@click.pass_context
def start(ctx, host: str, port: int):
    """Start a kiara (web) service."""

    from kiara_plugin.service.openapi.service import KiaraOpenAPIService

    try:
        import uvloop

        uvloop.install()
    except Exception:
        pass

    kiara_api: KiaraAPI = ctx.obj.kiara_api
    kiara_service = KiaraOpenAPIService(kiara_api=kiara_api)
    import uvicorn

    app = kiara_service.app()
    config = uvicorn.Config(app=app, host=host, port=port, log_level="info")
    server = uvicorn.Server(config=config)
    server.run()



# -*- coding: utf-8 -*-

"""This module contains the metadata (and other) models that are used in the ``kiara_plugin.service`` package.

Those models are convenience wrappers that make it easier for *kiara* to find, create, manage and version metadata -- but also
other type of models -- that is attached to data, as well as *kiara* modules.

Metadata models must be a sub-class of [kiara.metadata.MetadataModel][kiara.metadata.MetadataModel]. Other models usually
sub-class a pydantic BaseModel or implement custom base classes.
"""
from typing import Any, Union

from pydantic import Field

from kiara.models import KiaraModel


class InternalErrorModel(KiaraModel):
    """A model describing an internal server-side error."""

    _kiara_model_id = "instance.internal_error"

    @classmethod
    def from_exception(cls, exception: Exception, status: Union[int, None] = None):

        if status is None:
            status = 500

        return InternalErrorModel.construct(
            status=status, msg=str(exception), exception=None
        )

    status: int = Field(description="The status code code.")
    msg: str = Field(description="The error message.")
    exception: Any = Field(
        description="More details about the underlying error.", default=None
    )
